#include "scheduler.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#define LOCAL_DATA_FILE	1


#ifndef LOCAL_DATA_FILE
#define	FILE_FOLDER	".scheduler"
#endif
#define FILE_NAME	"scheduler.dat"
#define CLOSE_FILE	""
#define OPEN_RDONLY	"r"
#define OPEN_RDWR	"r+"
#define FIELD_SEPARATOR	'\f'

#define BUFFER_SIZE	1024
#define HEADER_SIZE	16
#define DATA_OFFSET	16



/******** External function declarations ********/

void init_task_queue();



/******** Internal function declarations ********/

static FILE *prepare_file(gchar *new, gchar *mode);
static void write_header(FILE *fp);
static gint get_directory(gchar **path);
static gint lookup_record(FILE *fp, gchar *date, gchar *hour,
                          gchar *minute, task_info *tip);
static gint insert_record(FILE *fin, FILE *fout, task_info *tip);
static gchar *right_trim(gchar *s);
static void format_digit2(gchar *buf, gint num);
static void set_record_date(gchar *buf, task_info *tip);
static void save_record(FILE *fp, gchar *date, task_info *tip);
static gchar *copy_string(gchar *buf, gchar *str, gint len);
static void substitute_char(gchar *s, gchar old, gchar new);



/******** External function definitions ********/

void read_task_for_day(GtkWidget *clist, task_info *tip)
{
   FILE *fp;
   gchar date[7], minute[3], hour[3];
   gint i, j;

   if((fp = prepare_file(FILE_NAME, OPEN_RDONLY)) == NULL)
      return;

   set_record_date(date, tip);

   /* get schedule data from file */
   fseek(fp, 0L, SEEK_SET);

   /* fix me: read time which is on the hour only at present */
   strcpy(minute, "00");

   for(i = HOUR_BEGIN, j = 0; i <= HOUR_END; i++, j++)
   {
      format_digit2(hour, i);

      if(lookup_record(fp, date, hour, minute, tip))
         gtk_clist_set_text(GTK_CLIST(clist), j, 1, tip->summary);
      else
      {
         fseek(fp, 0L, SEEK_SET);
         gtk_clist_set_text(GTK_CLIST(clist), j, 1, "");
      }
   }
}



gint save_current_task(task_info *tip)
{
   FILE *fin, *fout;
   gchar cmd[256], name[128], temp[128];

   /* Save only if summary and details are initialised */
   if((tip->summary == NULL) || (tip->details == NULL))
      return 0;

   strcpy(name, FILE_NAME);
   sprintf(temp, "/tmp/scheduler.%d", getpid());

   /* allow NULL in fin */
   fin = prepare_file(name, OPEN_RDWR);

   /* Must not use prepare_file to open the output file */
   if((fout = fopen(temp, "w+")) == NULL)
      return 0;

   write_header(fout);

   if((fin == NULL) || insert_record(fin, fout, tip))
   {
      /* Close the output file "temp" */
      fclose(fout);

      if(fin != NULL)
      {
         /* Close the input file "name" */
         prepare_file(name, CLOSE_FILE);
      }

      /* Remove temp file if replace output file failed */
      if(rename(temp, name) != 0)
      {
         /* Try again using system command */
         sprintf(cmd, "mv %s %s", temp, name);

         if(system(cmd) != 0)
            unlink(temp);
         else
         {
            init_task_queue();
            return 1;
         }
      }
      else
      {
         init_task_queue();
         return 1;
      }
   }

   return 0;
}



gboolean read_single_record(task_info *tip)
{
   FILE *fp;
   gchar hour[3], minute[3], date[7];

   if((fp = prepare_file(FILE_NAME, OPEN_RDONLY)) == NULL)
      return FALSE;

   set_record_date(date, tip);

   format_digit2(hour, tip->hour);
   format_digit2(minute, tip->minute);

   fseek(fp, 0L, SEEK_SET);

   if(lookup_record(fp, date, hour, minute, tip))
      return TRUE;
   else
      return FALSE;
}



gboolean save_single_record(task_info *tip)
{
   /* calling from details.c */

   return save_current_task(tip);
}



gboolean get_monthly_summary(task_info *tip, gint *fs)
{
   FILE *fp;
   gchar date[7], day[3], buf[BUFFER_SIZE];
   gint i, n;

   if((fp = prepare_file(FILE_NAME, OPEN_RDONLY)) == NULL)
      return FALSE;

   for(i = 0; i <= 31; i++)
      fs[i] = 0;

   fseek(fp, 0L, SEEK_SET);

   set_record_date(date, tip);

   date[4] = day[3] = '\0';

   while(fgets(buf, BUFFER_SIZE, fp) != NULL)
   {
      if(strncmp(buf, date, 4) == 0)
      {
         copy_string(day, buf + 4, 2);
         sscanf(day, "%d", &n);

         fs[n]++;
      }
   }

   return 1;
}



task_info *get_event_of_day(gint year, gint month, gint day)
{
   FILE *fp;
   task_info *head, *last, *p;
   gchar date[5], s1[5], s2[3], s3[3], buf[BUFFER_SIZE];
   gint len;

   if((fp = prepare_file(FILE_NAME, OPEN_RDONLY)) == NULL)
      return FALSE;

   head = last = NULL;

   format_digit2(s1, year);
   format_digit2(s2, month);
   format_digit2(s3, day);

   sprintf(date, "%s%s%s", s1 + 2, s2, s3);

   fseek(fp, 0L, SEEK_SET);

   while(fgets(buf, BUFFER_SIZE, fp) != NULL)
   {
      if(strncmp(buf, date, 6) == 0)
      {
         p = (task_info *)calloc(1, sizeof(task_info));

         if(head == NULL)
            head = p;
         else
            last->next = p;

         p->year = year;
         p->month = month;
         p->day = day;

         copy_string(s1, buf + 6, 2);
         copy_string(s2, buf + 8, 2);

         sscanf(s1, "%d", &p->hour);
         sscanf(s2, "%d", &p->minute);

         copy_string(s1, buf + 10, 2);
         sscanf(s1, "%d", &len);

         /* get alarm set off duration and unit */
         copy_string(s1, buf + 12, 2);
         sscanf(s1, "%d", &p->tsoff);
         p->tunit = buf[14];

         p->summary = g_malloc(len + 1);
         copy_string(p->summary, buf + DATA_OFFSET, len);

         last = p;
      }
   }

   return head;
}



/******** Internal function definitions ********/

static FILE *prepare_file(gchar *name, gchar *mode)
{
   static FILE *fp;
   static gchar *dir, *fmode, fname[128];
   gchar path[256];

   if(mode == CLOSE_FILE)
   {
      /* Close the opened file */
      fclose(fp);
      sprintf(name, "%s/%s", dir, fname);
      strcpy(fname, "");
      fp = NULL;
      return fp;
   }

   if(fp && (strcmp(fname, name) == 0))
   {
      /* Determine whether we need to reopen the file for writing */
      if((fmode != mode) && (mode == OPEN_RDWR))
      {
         fmode = CLOSE_FILE;
         strcpy(fname, "");
         fclose(fp);
      }
      else
      {
         /* access mode is ok so just reset i/o file beginning */
         fseek(fp, (long)HEADER_SIZE, SEEK_SET);
         return fp;
      }
   }

   if(dir == NULL)
      get_directory(&dir);

   /* Need to open the file so set the full path first */
   sprintf(path, "%s/%s", dir, name);

   if((fp = fopen(path, mode)) == NULL)
   {
      /* Create the file if it is in writing mode */
      if(mode == OPEN_RDWR && (fp = fopen(path, "w+")) != NULL)
         write_header(fp);
   }
   else
      fseek(fp, (long)HEADER_SIZE, SEEK_SET);

   if(fp != NULL)
   {
      fmode = mode;
      strcpy(fname, name);
   }

   return fp;
}



static void write_header(FILE *fp)
{
   /* go to the beginning of the file */
   fseek(fp, 0L, SEEK_SET);

   /* string written must be the same size of HEADER_SIZE */
   fprintf(fp, "%4d %10s\n", 0, "*RESERVED*");
}



static gint get_directory(gchar **rtn)
{
   gchar path[256];

#ifdef LOCAL_DATA_FILE
#ifdef PDA
   strcpy(path, "/buffer");
#else
   getcwd(path, 256);
#endif

   *rtn = g_strdup(path);

   return 1;
#else
   struct stat buf;
   gint f;

   sprintf(path, "%s/%s", getenv("HOME"), FILE_FOLDER);

   /* Allocate memory and copy the path to it */
   *rtn = g_strdup(path);

   if((f = stat(path, &buf)) != 0)
   {
      /* Folder not exists so create it */
      f = mkdir(path, 600);
   }

   return (f == 0);
#endif
}



static gint insert_record(FILE *fin, FILE *fout, task_info *tip)
{
   gchar *p, date[11], buf[BUFFER_SIZE];
   gint f, done;

   set_record_date(date, tip);

   p = date + 6;

   format_digit2(p, tip->hour);
   format_digit2(p + 2, tip->minute);

   done = 0;

   while(fgets(buf, BUFFER_SIZE, fin))
   {
      if(! done)
      {
         f = strncmp(buf, date, 10);

         /* don't save old record */
         if(f == 0)
            continue;

         if(f > 0)
         {
            /* got insertion point so save record now */
            save_record(fout, date, tip);
            done = 1;
         }
      }

      fputs(buf, fout);
   }

   /* If not done yet, save record at the end of the file */
   if(! done)
      save_record(fout, date, tip);

   write_header(fout);

   return 1;
}



static gint lookup_record(FILE *fp, gchar *date, gchar *hour,
                          gchar *minute, task_info *tip)
{
   gchar *p, tmp[20], buf[BUFFER_SIZE];
   gint f, len;

   f = 0;

   g_free(tip->summary);
   g_free(tip->details);

   len = sprintf(tmp, "%s%s%s", date, hour, minute);

   while(fgets(buf, BUFFER_SIZE, fp) != NULL)
   {
      if(strncmp(buf, tmp, len) == 0)
      {
         /* get length of scheduler summary text */
         copy_string(tmp, buf + len, 2);
         sscanf(tmp, "%d", &len);

         /* get alarm set off duration and unit */
         copy_string(tmp, buf + len + 2, 2);
         sscanf(tmp, "%d", &tip->tsoff);
         tip->tunit = buf[len + 3];

         p = buf + DATA_OFFSET;

         tip->summary = g_malloc(len + 1);
         copy_string(tip->summary, p, len);

         p += len;
         right_trim(p);

         len = strlen(p);
         tip->details = g_malloc(len + 1);
         copy_string(tip->details, p, len);

#if (BUG_FIXED==1)
         /* after reading, escaped newline characters must be *
          * restored. */
         substitute_char(tip->details, FIELD_SEPARATOR, '\n');
#else
         /* if newlines are restored, Text widget will cause nano-X *
          * crash, so now replace them with blanks */
         substitute_char(tip->details, FIELD_SEPARATOR, ' ');
#endif

         return 1;
      }
   }

   tip->summary = g_strdup("");
   tip->details = g_strdup("");
   tip->tsoff = DEF_SETOFF_TM;
   tip->tunit = DEF_TIME_UNIT;

   return 0;
}


static gchar *right_trim(gchar *s)
{
   gchar *p;

   p = s + strlen(s) - 1;

   while(p >= s) 
   {
      if(isspace(*p) || *p == '\n')
      {
         *p = '\0';
         break;
      }

      p--;
   }

   return s;
}



static void format_digit2(gchar *buf, gint num)
{
   if(num < 10)
      sprintf(buf, "0%d", num);
   else
      sprintf(buf, "%d", num);
}



static void set_record_date(gchar *buf, task_info *tip)
{
   gchar year[5], month[3], day[3];

   format_digit2(year, tip->year);
   format_digit2(month, tip->month);
   format_digit2(day, tip->day);

   sprintf(buf, "%s%s%s", year + 2, month, day);
}



static void save_record(FILE *fp, gchar *date, task_info *tip)
{
   gchar *s;

   /* no summary means removing record only */
   if(tip->summary && *tip->summary)
   {

      /* Write header to the output file */
      fprintf(fp, "%s%2d%2d%c ", date, strlen(tip->summary),
              tip->tsoff, tip->tunit);

      /* before saving, newline characters must be escaped */
      s = g_strdup(tip->details);
      substitute_char(s, '\n', FIELD_SEPARATOR);

      /* Write data which terminates by a newline char. */
      fprintf(fp, "%s%s\n", tip->summary, s);

      g_free(s);
   }
}



static gchar *copy_string(gchar *buf, gchar *str, gint len)
{
   strncpy(buf, str, len);

   buf[len] = '\0';

   return buf;
}



static void substitute_char(gchar *s, gchar old, gchar new)
{
   while(*s)
   {
      if(*s == old)
         *s = new;

      s++;
   }
}



