#include "scheduler.h"
#include <stdlib.h>
#include <unistd.h>
#if (USE_SYS_ALARM == 1)
#include <signal.h>
#endif


#ifdef PCX
#define TRANSIENT_WIN
#endif

#define MODAL_POPUP     	1
#define DESTROY_ON_CLOSE	1


#define POPUP_WD		(TOPLEVEL_WD - 60)
#define POPUP_HT		(TOPLEVEL_HT - 160)
#define POPUP_X			(TOPLEVEL_X + 30)
#define POPUP_Y			(TOPLEVEL_Y + 80)

#if (USE_SYS_ALARM == 1)
#define ONE_SECOND		1
#else
#define ONE_SECOND		1000
#endif

#define ONE_MINUTE		(ONE_SECOND * 60)
#define FIVE_MINUTES		(ONE_MINUTE * 5)
#define ONE_QUARTER		(ONE_MINUTE * 15)
#define ONE_HOUR		(ONE_HOUR * 60)

#ifndef CHECK_INTERVAL
#define	CHECK_INTERVAL		ONE_MINUTE
#endif

#define SNOOZE_TIME		5	// five minutes


static gchar *MonthName[] = { "",
                              "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                              "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static GtkWidget *dialog;
static task_info *head, *curr;
static gint year, month, day;
static gint snooze_c, snooze_t;



/******** External function declarations ********/

task_info *get_event_of_day(gint year, gint month, gint day);
GtkWidget *get_toplevel_widget();
void jump_to_date(time_t tref);



/******** Internal function declarations ********/

static gint check_task_queue();
static time_t get_current_time(gint *y, gint *m, gint *d,
                               gint *hh, gint *mm);
static char *format_digit2(gchar *buf, gint num);
static void show_message(gchar *msg, gint snooze);
static void show_message_window(gint flg);
static void snooze_schedule(GtkWidget *widget, GdkEvent *event,
                            gpointer data);
static void close_message();
static void reset_schedule(GtkWidget *widget, GdkEvent *event,
                           gpointer data);
#if (USE_SYS_ALARM == 1)
static void alarm_handler();
#endif



/******** External function definitions ********/

void init_task_queue()
{
   static guint tid;
   gint y, m, d, hh, mm;

   /* free existing task queue structure first */
   curr = head;

   while(curr)
   {
      head = curr->next;
      free(curr);
      curr = head;
   }

   /* Get events of the day */
   get_current_time(&y, &m, &d, &hh, &mm);

   year = y;
   month = m;
   day = d;

   head = curr = get_event_of_day(y, m, d);

   /* ignore passed events, get to the nearest future event */
#if (DEBUG==1)
   printf("current time is %d/%d/%d %d:%d\n", y, m, d, hh, mm);
   printf("removing expired events from queue:\n");
#endif

   while(curr && curr->hour <= hh && curr->minute < mm)
   {
#if (DEBUG==1)
      gchar s1[3], s2[3];
      printf("   %s:%s - \"%s\".\n",
             format_digit2(s1, curr->hour),
             format_digit2(s2, curr->minute),
             curr->summary);
#endif
      curr = curr->next;
      g_free(head->summary);
      free(head);
      head = curr;
   }

#if (USE_SYS_ALARM == 1)
   /* remove previously installed timeout callback first */
   if(tid != 0)
      alarm(0);

   /* do an initial check on scheduled task */
   check_task_queue();

   /* add alarm for subsequent task checking */
   tid = alarm(CHECK_INTERVAL);
   signal(SIGALRM, alarm_handler);
#else
   /* remove previously installed timeout callback first */
   if(tid != 0)
      gtk_timeout_remove(tid);

   /* do an initial check on scheduled task */
   check_task_queue();

   /* add timeout callback for subsequent task checking */
   tid = gtk_timeout_add(CHECK_INTERVAL,
                         (GtkFunction)check_task_queue,
                         (gpointer)NULL);
#endif
}



/******** Internal function definitions ********/

static gint check_task_queue()
{
   time_t ts;
   gchar s1[3], s2[3], s3[30], msg[256];
   gint y, m, d, hh, mm, tnow, tset, tled, toff;

   ts = get_current_time(&y, &m, &d, &hh, &mm);

   /* reinitialise task queue if date changed */
   if(y != year || m != month || d != day)
   {
      jump_to_date(ts);
      init_task_queue();
   }

   if(curr != NULL)
   {
      /* convert to minute r.t. the beginning of the day */
      tset = (curr->day * 24 + curr->hour) * 60 + curr->minute;
      tnow = (d * 24 + hh) * 60 + mm;
      tled = tset - tnow;

      switch(curr->tunit)
      {
         case TIME_MINUTE: toff = curr->tsoff; break;
         case TIME_HOUR  : toff = curr->tsoff * 60; break;
         case TIME_DAY   : toff = curr->tsoff * 24 * 60; break;
         default         : toff = DEF_SETOFF_TM;
      }

#if (DEBUG==1)
      printf("%d: tled=%d toff=%d\n", getpid(), tled, toff);
#endif

      /* set off alarm if current time falls in trigger region */
      if(tled >= 0)
      {
         /* no snooze if set time arrived */
         if(tled > 0 && snooze_t > 0)
         {
#if (DEBUG==1)
            printf("%d: snooze for %d min(s).\n", getpid(), snooze_t);
#endif
            snooze_t--;
            return 1;
         }

         if(tled <= toff)
         {
            if(tled >= 60)
            {
               sprintf(msg, "%s %2d, %d   %s:%s\n \n\"%s\".",
                       MonthName[curr->month], curr->day, curr->year,
                       format_digit2(s1, curr->hour),
                       format_digit2(s2, curr->minute),
                       curr->summary);
            }
            else
            {
               if(tled == 0)
               {
                  gdk_beep(); 
                  strcpy(s3, "now");
               }
               else if(tled == 1)
                  sprintf(s3, "%d min. left", tled); 
               else
                  sprintf(s3, "%d mins. left", tled);

               sprintf(msg, "At  %s:%s  (%s)\n \n\"%s\"",
                       format_digit2(s1, curr->hour),
                       format_digit2(s2, curr->minute),
                       s3, curr->summary);
            }

            show_message(msg, (tnow < tset));
            gdk_beep();
         }
      }
      else
      {
         /* remove expired task and move to the next event */
         /* midnight crossing may not work here !!! */
         if(tnow > tset)
         {
#if (DEBUG==1)
            printf("reset alarm at tset=%d tnow=%d\n", tset, tnow);
#endif
            reset_schedule(NULL, NULL, NULL);
         }

#if (DEBUG==1)
         if(curr != NULL && tset - tnow > 0)
            printf("pending event: %s:%s - \"%s\", %d min(s) left.\n",
                   format_digit2(s1, curr->hour),
                   format_digit2(s2, curr->minute),
                   curr->summary, tset - tnow);
#endif
      }
   }

#if (DEBUG==1)
   if(curr == NULL)
      printf("no pending event for today\n");
#endif

   /* return non-zero to repeat the time */
   return 1;
}



static time_t get_current_time(gint *y, gint *m, gint *d,
                             gint *hh, gint *mm)
{
   time_t now;
   gchar buf[40];

   time(&now);

   strftime(buf, sizeof buf, "%Y %m %e %k %M", localtime(&now));

   sscanf(buf, "%d %d %d %d %d", y, m, d, hh, mm);

   return now;
}



static gchar *format_digit2(gchar *buf, gint num)
{
   if(num < 10)
      sprintf(buf, "0%d", num);
   else
      sprintf(buf, "%d", num);

   return buf;
}



static void show_message(gchar *msg, gint snooze)
{
   static GtkWidget *label, *snooze_button, *ok_button;

   /* avoid creating multiple popups for the same message */
   if(dialog == NULL)
   {
#ifdef TRANSIENT_WIN
      GtkWidget *parent;
#endif

      /* Create the widgets */
      dialog = gtk_dialog_new();
      gtk_widget_set_uposition(GTK_WIDGET(dialog), POPUP_X, POPUP_Y);
      gtk_widget_set_usize(GTK_WIDGET(dialog), POPUP_WD, POPUP_HT);

      /* if newlines embedded, break message into multiple lines */
      label = gtk_label_new(msg);
      gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), label);

      snooze_button = gtk_button_new_with_label("Snooze");
      gtk_signal_connect_object(GTK_OBJECT(snooze_button), "clicked",
                                GTK_SIGNAL_FUNC(snooze_schedule),
                                (gpointer)NULL);
      gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
                        snooze_button);

      ok_button = gtk_button_new_with_label("OK");
      gtk_signal_connect_object(GTK_OBJECT(ok_button), "clicked",
                                GTK_SIGNAL_FUNC(reset_schedule),
                                (gpointer)NULL);
      gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
                        ok_button);

      /* Add the label, and show everything we've added to the dialog. */
      gtk_widget_show_all(dialog);

#ifdef TRANSIENT_WIN
      parent = get_toplevel_widget();
      gdk_window_set_transient_for(dialog->window, parent->window);
#endif
   }
   else
      gtk_label_set_text(GTK_LABEL(label), msg);

   if(snooze)
      gtk_widget_map(snooze_button);
   else
      gtk_widget_unmap(snooze_button);

   show_message_window(1);
}



static void show_message_window(gint flg)
{
   if(dialog == NULL)
      return;

   if(flg)
   {
#if (DEBUG==1)
      printf("%d: map message window\n", getpid());
#endif

      gtk_widget_map(dialog);

#if (MODAL_POPUP==1)
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
#endif

      gdk_window_raise(dialog->window);
   }
   else
   {
#if (DEBUG==1)
      printf("%d: unmap message window\n", getpid());
#endif

#if (MODAL_POPUP==1)
      gtk_window_set_modal(GTK_WINDOW(dialog), FALSE);
#endif

      gtk_widget_unmap(dialog);
   }
}



static void snooze_schedule(GtkWidget *widget, GdkEvent *event,
                            gpointer data)
{
   snooze_c = snooze_t = SNOOZE_TIME;

   close_message();
}



static void close_message()
{
#if (DESTROY_ON_CLOSE==1)
   gtk_widget_destroy(dialog);

   dialog = NULL;
#else
   show_message_window(0);
#endif
}



static void reset_schedule(GtkWidget *widget, GdkEvent *event,
                           gpointer data)
{
   if(dialog != NULL)
      close_message(NULL, NULL, NULL);

#if (DEBUG==1)
   {
      gchar s1[3], s2[3];
      printf("removing %s:%s - \"%s\" from queue.\n",
             format_digit2(s1, curr->hour),
             format_digit2(s2, curr->minute),
             curr->summary);
   }
#endif

   /* remove current task and set up for the next one */
   curr = curr->next;
   g_free(head->summary);
   free(head);
   head = curr;

#if (DEBUG==1)
   if(curr != NULL)
   {
      gchar s1[3], s2[3];
      printf("next event: %s:%s - \"%s\".\n",
             format_digit2(s1, curr->hour),
             format_digit2(s2, curr->minute),
             curr->summary);
   }
   else
      printf("no next event now\n");
#endif

   snooze_c = snooze_t = 0;
}


#if (USE_SYS_ALARM == 1)
static void alarm_handler()
{
#if (DEBUG==1)
   printf(".");
   fflush(stdout);
#endif

   check_task_queue();

   /* must set alarm again */
   alarm(CHECK_INTERVAL);
}
#endif



