/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include <config.h>

/* gcc -ansi -pedantic on GNU/Linux causes warnings and errors
 * unless this is defined:
 * warning: #warning "Files using this header must be compiled with _SVID_SOURCE or _XOPEN_SOURCE"
 */
#ifndef _XOPEN_SOURCE
#  define _XOPEN_SOURCE 1
#endif

#include <stdlib.h>
#include <sys/types.h>

#if defined (HAVE_IPC_H) && defined (HAVE_SHM_H) && defined (HAVE_XSHM_H)
#define USE_SHM
#endif


#include "gdk.h"
#include "gdkprivate.h"

extern GR_SCREEN_INFO gdk_screen_info;
void gdk_image_put_normal (GdkDrawable *drawable,
				  GdkGC       *gc,
				  GdkImage    *image,
				  gint         xsrc,
				  gint         ysrc,
				  gint         xdest,
				  gint         ydest,
				  gint         width,
				  gint         height);
static void gdk_image_put_shared (GdkDrawable *drawable,
				  GdkGC       *gc,
				  GdkImage    *image,
				  gint         xsrc,
				  gint         ysrc,
				  gint         xdest,
				  gint         ydest,
				  gint         width,
				  gint         height);


static GList *image_list = NULL;

static GR_GC_ID  gc_for_image = 0;
#define ENSURE_GC if (!gc_for_image){\
                 gc_for_image= GrNewGC();\
                 GrSetGCUseBackground(gc_for_image, GR_FALSE);\
                }

#define PRINTFILE(s) g_message("gdkimage.c:%s()",s)

void
gdk_image_exit (void)
{
  GdkImage *image;

  while (image_list)
    {
      image = image_list->data;
      gdk_image_destroy (image);
    }
}

GdkImage *
gdk_image_new_bitmap(GdkVisual *visual, gpointer data, gint w, gint h)
/*
 * Desc: create a new bitmap image
 */
{
        GdkImage *image;
        GdkImagePrivate *private;
	PRINTFILE("gdk_image_new_bitmap");

        private = g_new(GdkImagePrivate, 1);
        image = (GdkImage *) private;
        private->image_put = gdk_image_put_normal;
        image->type = GDK_IMAGE_NORMAL;
        image->visual = visual;
        image->width = w;
        image->height = h;
        image->depth = 1;
	private->ximage = GrNewPixmap(w,h, data);
        image->byte_order = LSBFirst;
        image->mem =  data;
        image->bpl = w << 3;
        image->bpp = 1;
        return(image);

} /* gdk_image_new_bitmap() */

static int
gdk_image_check_xshm()
/* 
 * Desc: query the server for support for the MIT_SHM extension
 * Return:  0 = not available
 *          1 = shared XImage support available
 *          2 = shared Pixmap support available also
 */
{
#if 0 //USE_SHM
  int major, minor, ignore;
  Bool pixmaps;
  
  if (XQueryExtension(display, "MIT-SHM", &ignore, &ignore, &ignore)) 
    {
      if (XShmQueryVersion(display, &major, &minor, &pixmaps )==True) 
	{
	  return (pixmaps==True) ? 2 : 1;
	}
    }
#endif /* USE_SHM */
  return 0;
}

void
gdk_image_init (void)
{
  if (gdk_use_xshm)
    {
      if (!gdk_image_check_xshm ())
	{
	  gdk_use_xshm = False;
	}
    }
}

GdkImage*
gdk_image_new (GdkImageType  type,
	       GdkVisual    *visual,
	       gint          width,
	       gint          height)
{
  GdkImage *image;
  GdkImagePrivate *private;

  //PRINTFILE("__IMPLEMENTED__ gdk_image_new");
  private = g_new (GdkImagePrivate, 1);
  image = (GdkImage*) private;

  private->image_put = gdk_image_put_normal;
  private->ximage = GrNewPixmap(width, height, NULL);
  private->bnImageInServer = GR_FALSE;
  image->type = GDK_IMAGE_NORMAL;
  image->visual = visual;
  image->width = width;
  image->height = height;
  image->depth = visual->depth;

  image->byte_order = LSBFirst; 
  if(gdk_screen_info.bpp < 8) image->bpl = width;
  else
  if(gdk_screen_info.bpp == 12) image->bpl = width*2;//((width+1)*3)/2;
  else
  	image->bpl = (width *gdk_screen_info.bpp) >> 3;
  image->mem = g_malloc(image->bpl * height);  
  image->bpp = (gdk_screen_info.bpp + 7) >> 3;
  return image;
}

GdkImage*
gdk_image_get (GdkWindow *window,
	       gint       x,
	       gint       y,
	       gint       width,
	       gint       height)
{
  GdkImage *image;
  GdkImagePrivate *private;
  GdkWindowPrivate *win_private;

  g_return_val_if_fail (window != NULL, NULL);

	//PRINTFILE("__IMPLEMENTED__ gdk_image_get");

  win_private = (GdkWindowPrivate *) window;
  if (GDK_DRAWABLE_DESTROYED(window)) return NULL;

  private = g_new (GdkImagePrivate, 1);
  image = (GdkImage*) private;

  private->image_put = gdk_image_put_normal;
  private->ximage = GrNewPixmap(width, height, NULL);
  private->bnImageInServer = GR_TRUE;
   
  image->type = GDK_IMAGE_NORMAL;
  image->visual = gdk_window_get_visual (window);
  image->width = width;
  image->height = height;
  image->depth = gdk_screen_info.bpp;
  /* the bpp return from server is sizeof(GR_PIXELVAL) , each pixel is
     store in a GR_PIXELVAL format */
  image->bpl = width *sizeof(GR_PIXELVAL);
  image->mem = g_malloc(image->bpl*height);
  image->bpp =  sizeof(GR_PIXELVAL) ;
  image->byte_order = LSBFirst;

  ENSURE_GC;
  GrCopyArea(private->ximage, gc_for_image, 0, 0,
        width, height, GDK_DRAWABLE_XID (window), x, y, MWROP_SRCCOPY);

  GrReadArea(private->ximage, 0, 0, width, height, image->mem);
  return image;
}

guint32
gdk_image_get_pixel (GdkImage *image,
		     gint x,
		     gint y)
{
  guint32 pixel;
  GR_PIXELVAL pixelval;
  GdkImagePrivate *private;

  g_return_val_if_fail (image != NULL, 0);

  private = (GdkImagePrivate *)image;
  GrReadArea(private->ximage, x, y, 1, 1, &pixelval);

  pixel = (guint32)pixelval;
  return pixel;

}

void
gdk_image_put_pixel (GdkImage *image,
		     gint x,
		     gint y,
		     guint32 pixel)
{
  GdkImagePrivate *private;
  GR_PIXELVAL pixelval;

	//PRINTFILE("gdk_image_put_pixel");
  g_return_if_fail (image != NULL);

  private = (GdkImagePrivate *) image;
  private->bnImageInServer = GR_TRUE;
  pixelval = (GR_PIXELVAL)pixel;
  ENSURE_GC;
  GrArea(private->ximage, gc_for_image, x, y, 1, 1, &pixelval,gdk_screen_info.pixtype); 
}

void
gdk_image_destroy (GdkImage *image)
{
	GdkImagePrivate *image_private;

	//PRINTFILE("gdk_image_destroy");
  	g_return_if_fail (image != NULL);
	image_private = (GdkImagePrivate*)image;
	
	if(image_private->ximage)
		GrDestroyWindow(image_private->ximage);
	if (image->mem)
		g_free(image->mem);
	g_free(image);
}

#if 0
void
gdk_image_put_buffer (GdkDrawable *drawable,
                      GdkGC       *gc,
                      guchar      *buffer,
                      gint         xsrc,
                      gint         ysrc,
                      gint         xdest,
                      gint         ydest,
                      gint         width,
                      gint         height,
                      gint         src_width,
                      gint         src_height)
{
        guchar * small_buffer, *ptr;
        gint h;
        g_return_if_fail (drawable != NULL);
        g_return_if_fail (buffer != NULL);
        g_return_if_fail (gc != NULL);

        if (GDK_DRAWABLE_DESTROYED(drawable)) return;

        ptr = small_buffer = (guchar *)g_malloc(width*height);

        buffer += src_width * ysrc + xsrc;

        h = height;
        while (h--)
        {
                memcpy(small_buffer,buffer,width);
                buffer +=src_width;
                small_buffer +=width;
        }
        GrArea(GDK_DRAWABLE_XID(drawable), GDK_GC_XGC(gc), xdest, ydest,
        width, height, ptr, gdk_screen_info.pixtype);

        g_free(ptr);
}
#else
void
gdk_image_put_buffer (GdkDrawable *drawable,
                      GdkGC       *gc,
                      guchar      *buffer,
                      gint         xsrc,
                      gint         ysrc,
                      gint         xdest,
                      gint         ydest,
                      gint         width,
                      gint         height,
                      gint         src_width,
                      gint         src_height,
		      gint	   alpha)
{
        guchar * small_buffer, *ptr;
        gint h;
        g_return_if_fail (drawable != NULL);
        g_return_if_fail (buffer != NULL);
        g_return_if_fail (gc != NULL);

        if (GDK_DRAWABLE_DESTROYED(drawable)) return;

	/* Support for 4bpp alpha. For nano-X, type is palette */
        if (gdk_screen_info.pixtype == MWPF_PALETTE)
        {
        	if (alpha)
                	GrAreaSrc(GDK_DRAWABLE_XID(drawable), GDK_GC_XGC(gc), xdest, ydest,
                width, height, xsrc,ysrc,src_width,buffer, MWPF_GRAY_PACKED_4_ALPHA,-1);
        	else
                	GrAreaSrc(GDK_DRAWABLE_XID(drawable), GDK_GC_XGC(gc), xdest, ydest,
                width, height, xsrc,ysrc,src_width,buffer, MWPF_GRAY_PACKED_4,-1);
        }
        else
        if (gdk_screen_info.pixtype == MWPF_TRUECOLOR444)
        {
        	GrAreaSrc(GDK_DRAWABLE_XID(drawable), GDK_GC_XGC(gc), xdest, ydest,
        width, height, xsrc,ysrc,src_width,buffer, MWPF_TRUECOLOR444_PACKED, alpha);
	}
	else
        	GrAreaSrc(GDK_DRAWABLE_XID(drawable), GDK_GC_XGC(gc), xdest, ydest,
        width, height, xsrc,ysrc,src_width,buffer, gdk_screen_info.pixtype, alpha);
}
#endif

void
gdk_image_put_normal (GdkDrawable *drawable,
		      GdkGC       *gc,
		      GdkImage    *image,
		      gint         xsrc,
		      gint         ysrc,
		      gint         xdest,
		      gint         ydest,
		      gint         width,
		      gint         height)
{
  GdkImagePrivate *image_private;

  g_return_if_fail (drawable != NULL);
  g_return_if_fail (image != NULL);
  g_return_if_fail (gc != NULL);

  if (GDK_DRAWABLE_DESTROYED(drawable)) return;
  image_private = (GdkImagePrivate*) image;

  g_return_if_fail (image->type == GDK_IMAGE_NORMAL);

  if(1)
//!image_private->bnImageInServer) 
  {
  	GrArea(image_private->ximage, GDK_GC_XGC(gc), 0, 0, 
		image->width, image->height, image->mem, gdk_screen_info.pixtype); 
  }
  GrCopyArea(GDK_DRAWABLE_XID(drawable), GDK_GC_XGC(gc), xdest, ydest,
	width, height, image_private->ximage, xsrc, ysrc, MWROP_SRCCOPY); 

}

static void
gdk_image_put_shared (GdkDrawable *drawable,
		      GdkGC       *gc,
		      GdkImage    *image,
		      gint         xsrc,
		      gint         ysrc,
		      gint         xdest,
		      gint         ydest,
		      gint         width,
		      gint         height)
{
#if 0 //USE_SHM
  GdkWindowPrivate *drawable_private;
  GdkImagePrivate *image_private;
  GdkGCPrivate *gc_private;

  g_return_if_fail (drawable != NULL);
  g_return_if_fail (image != NULL);
  g_return_if_fail (gc != NULL);

  drawable_private = (GdkWindowPrivate*) drawable;
  if (drawable_private->destroyed)
    return;
  image_private = (GdkImagePrivate*) image;
  gc_private = (GdkGCPrivate*) gc;

  g_return_if_fail (image->type == GDK_IMAGE_SHARED);

  XShmPutImage (drawable_private->xdisplay, drawable_private->xwindow,
		gc_private->xgc, image_private->ximage,
		xsrc, ysrc, xdest, ydest, width, height, False);
#else /* USE_SHM */
  g_error ("trying to draw shared memory image when gdk was compiled without shared memory support");
#endif /* USE_SHM */
}
